/*******************************************************************************
*
* E M B E D D E D   W I Z A R D   P R O J E C T
*
*                                                Copyright (c) TARA Systems GmbH
*                                    written by Paul Banach and Manfred Schweyer
*
********************************************************************************
*
* This software is delivered "as is" and shows the usage of other software
* components. It is provided as an example software which is intended to be
* modified and extended according to particular requirements.
*
* TARA Systems hereby disclaims all warranties and conditions with regard to the
* software, including all implied warranties and conditions of merchantability
* and non-infringement of any third party IPR or other rights which may result
* from the use or the inability to use the software.
*
********************************************************************************
*
* DESCRIPTION:
*   This file is part of the interface (glue layer) between an Embedded Wizard
*   generated UI application and the board support package (BSP) of a dedicated
*   target.
*   This template is responsible to initialize the display hardware of the board
*   and to provide the necessary access to update the display content.
*   The color format of the framebuffer has to correspond to the color format
*   of the Graphics Engine.
*
*   Important: This file is intended to be used as a template. Please adapt the
*   implementation according your particular hardware.
*
*******************************************************************************/

#include <string.h>

#include "ewrte.h"
#include "ewgfx.h"

#include "ew_bsp_display.h"
#include "gfx_system.h"


/*******************************************************************************
* FUNCTION:
*   EwBspDisplayInit
*
* DESCRIPTION:
*   The function EwBspDisplayInit initializes the display hardware and returns
*   the display parameter.
*
* ARGUMENTS:
*   aGuiWidth,
*   aGuiHeight   - Size of the GUI in pixel.
*   aDisplayInfo - Display info data structure.
*
* RETURN VALUE:
*   Returns 1 if successful, 0 otherwise.
*
*******************************************************************************/
int EwBspDisplayInit( int aGuiWidth, int aGuiHeight, XDisplayInfo* aDisplayInfo )
{
  int              guiWidth;
  int              guiHeight;
  int              viewportX;
  int              viewportY;
  int              viewportWidth;
  int              viewportHeight;
  int              width;
  int              height;
  XGfxSystemInfo   gfxSystemInfo;

  /* check and clean display info structure */
  if ( !aDisplayInfo )
    return 0;

  memset( aDisplayInfo, 0, sizeof( XDisplayInfo ));

  /* get gfx system info structure */
  GfxSystemGetInfo( &gfxSystemInfo );

  if ( gfxSystemInfo.Type != GFX_SYSTEM_DRM )
  {
    EwPrint( "Invalid GFX system\n" );
    return 0;
  }

  width  = gfxSystemInfo.Width;
  height = gfxSystemInfo.Height;


  /* adjust size of gui if viewport is rotated */
  #if (( EW_VIEWPORT_ROTATION == 90 ) || ( EW_VIEWPORT_ROTATION == 270 ))
    guiWidth  = aGuiHeight;
    guiHeight = aGuiWidth;
  #else
    guiWidth  = aGuiWidth;
    guiHeight = aGuiHeight;
  #endif

  #if EW_VIEWPORT_AUTOSCALE == 1

    if (((float)width / (float)guiWidth ) < ((float)height / (float)guiHeight ))
    {
      /* modify height of viewport to keep the aspect ratio */
      viewportWidth  = width;
      viewportHeight = guiHeight * width / guiWidth;
      viewportX      = 0;
      viewportY      = ( height - viewportHeight ) / 2;
    }
    else
    {
      /* modify width of viewport to keep the aspect ratio */
      viewportWidth  = guiWidth * height / guiHeight;
      viewportHeight = height;
      viewportX      = ( width - viewportWidth ) / 2;
      viewportY      = 0;
    }

  #else

    viewportWidth  = guiWidth;
    viewportHeight = guiHeight;
    viewportX      = ( width  - guiWidth ) / 2;
    viewportY      = ( height - guiHeight ) / 2;

  #endif

  /* return the current display configuration */
  aDisplayInfo->FrameBuffer    = (void*)(intptr_t)gfxSystemInfo.FrameBuffer;
  aDisplayInfo->DisplayWidth   = width;
  aDisplayInfo->DisplayHeight  = height;
  aDisplayInfo->ViewportX      = viewportX;
  aDisplayInfo->ViewportY      = viewportY;
  aDisplayInfo->ViewportWidth  = viewportWidth;
  aDisplayInfo->ViewportHeight = viewportHeight;
  aDisplayInfo->Name           = ( char* )gfxSystemInfo.Interface;

  return 1;
}


/*******************************************************************************
* FUNCTION:
*   EwBspDisplayDone
*
* DESCRIPTION:
*   The function EwBspDisplayDone deinitializes the display hardware.
*
* ARGUMENTS:
*   aDisplayInfo - Display info data structure.
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspDisplayDone( XDisplayInfo* aDisplayInfo )
{
}


/*******************************************************************************
* FUNCTION:
*   EwBspDisplayUpdate
*
* DESCRIPTION:
*   The function EwBspDisplayUpdate updates the screen content.
*
* ARGUMENTS:
*   aContext - Display context.
*   aArea    - The update area.
*
* RETURN VALUE:
*   None
*
*******************************************************************************/
void EwBspDisplayUpdate( void* aContext, XRect aArea )
{
  /* start an update if there is anything to draw */
  if ( !EwIsRectEmpty( aArea ))
    GfxSystemUpdateSurface( aContext,
      aArea.Point1.X, aArea.Point1.Y, aArea.Point2.X, aArea.Point2.Y );
}


/* mli, msy */
